
SuperStrict

Rem
bbdoc: Marching squares algorithm implementation
about: This module provides an implementation of the marching squares algorithm. <br />
Short info: The marching squares algorithm is an algorithm to plot implicit functions. <br />
Working with this module is kind of straightforward: At first, you must define a line drawing function for the module to use. <br />
This Function has to accept the following parameters: The rendertarget (an Object), the four coordinates of the line (x and y coordinates of <br />
starting- and endpoint) (Integers) and the color of the line (an Integer). <br />
Create an object of TCurvePlotter and feed it with the necessary information: <br />
-The sampling function, the sample's width and height<br />
-Whether or not you want to interpolate<br />
-the threshold (look up how the algorithm works if you want to know what the threshold is)<br />
-the number of cells in horizontal and vertical direction<br />
-the color of the contour which will be plotted<br /><br />
After that, invoke the "Plot"-Method, which will use your line drawing function. For plotting you can also specify a viewport.
End Rem
Module Leiter.Marchingsquares

ModuleInfo "Author: Calvin 'n-Halbleiter' Bornhofen"
ModuleInfo "Version: 1.0"
ModuleInfo "License: Public Domain"
ModuleInfo "Copyright: Calvin Bornhofen"

'Imported modules
Import BRL.Blitz

Rem
bbdoc: Curve plotter Type
End Rem
Type TCurvePlotter
	
	Rem
	References for implementation:
		Interactive Computer Graphics (5th edition):
			p. 89 for image of cell types
			pp. 86-92 for algorithm description
		Wikipedia:
			Article "Linear_Interplation" for weighted interpolation (used at edges)
	End Rem
	
	Field _SampleFnc:Float(X:Int,Y:Int)
	Field _SampleWidth:Int
	Field _SampleHeight:Int
	Field _fInterpolate:Int
	Field _Threshold:Float
	Field _CellNum:Int
	Field _ColorARGB:Int
	
	Field _LineFnc(Target:Object,X1:Int,Y1:Int,X2:Int,Y2:Int,ARGB:Int)
	Rem
	bbdoc: Line function setter
	about: Use this method to set the used line drawing function to your own
	End Rem
	Method SetLineFnc(Fnc(Target:Object,X1:Int,Y1:Int,X2:Int,Y2:Int,ARGB:Int))
		Self._LineFnc=Fnc
	End Method
	
	Rem
	bbdoc: Set interpolation
	about: Set whether the plotting will be interpolated or not
	End Rem
	Method SetInterpolation(fInterpolate:Int=False)
		Self._fInterpolate=fInterpolate
	End Method
	
	Rem
	bbdoc: Set the threshold
	about: Use this to set the used threshold
	End Rem
	Method SetThreshold(Threshold:Float=0)
		Self._Threshold=Threshold
	End Method
	
	Rem
	bbdoc: Set cell number
	about: Use this to set the number of cells in horizontal and vertical direction
	End Rem
	Method SetCellNum(Size:Int)
		Self._CellNum=Size
	End Method
	
	Rem
	bbdoc: Set color
	about: Use this to set the color (as an ARGB integer) of the plotted contour
	End Rem
	Method SetColorARGB(ARGB:Int)
		Self._ColorARGB=ARGB
	End Method
	
	Rem
	bbdoc: Set sample function
	about: Use this to set the function which returns the used data
	End Rem
	Method SetSampleFnc(Fnc:Float(X:Int,Y:Int))
		Self._SampleFnc=Fnc
	End Method
	
	Rem
	bbdoc: Set sample size
	about: Use this to set the maximum size of the sample (limits the viewport of plotting)
	End Rem
	Method SetSampleSize(Width:Int,Height:Int)
		Self._SampleWidth=Width
		Self._SampleHeight=Height
	End Method
	
	'VX1,VY1,VX2,VY2: The viewport of the data which will be sampled
	Rem
	bbdoc: Plot
	about: This function plots to the specified rendertarget
	End Rem
	Method Plot(Target:Object,VX1:Int=0,VY1:Int=0,VX2:Int=-1,VY2:Int=-1)
		If ((Self._SampleFnc<>Null) And (Self._CellNum<>0))
			'Viewport correction
			If ((VX1<0) Or (VX1>Self._SampleWidth)) Then VX1=0
			If ((VY1<0) Or (VY1>Self._SampleHeight)) Then VY1=0
			If ((VX2<0) Or (VX2>Self._SampleWidth)) Then VX2=Self._SampleWidth
			If ((VY2<0) Or (VY2>Self._SampleHeight)) Then VY2=Self._SampleHeight
			Local VW:Int=Abs(VX2-VX1),VH:Int=Abs(VY2-VY1)
			
			'Setting of environment variables & Precalculations
			Local CellWidth:Int=Int(Double(VW)/Double(Self._CellNum)),CellHeight:Double=Int(Double(VH)/Double(Self._CellNum))
			
			Self.___Target=Target
			
			Local N:Int,iX:Int,iY:Int
			'March all cells
			For iX=0 Until Self._CellNum
				For iY=0 Until Self._CellNum
					Local X:Int,Y:Int
					X=VX1+iX*CellWidth
					Y=VY1+iY*CellHeight
					
					Self.__MarchCell(X,Y,CellWidth,CellHeight)
				Next
			Next
			
			Self.___Target=Null
		End If
	End Method
	Field ___Target:Object
	
	Rem
	bbdoc: Private method. DO NOT TOUCH!
	End Rem
	Method __MarchCell(X:Int,Y:Int,CellWidth:Int,CellHeight:Int)
		Rem
		fLine: Whether a line is to draw or not
		LX1,LY1,LX2,LY2: Coordinate of endpoints of line
		UL,UR,LR,LL: Values of sample at edges (Upper Left, Upper Right, Lower Right, Lower Left)
		N: Cell type
		End Rem
		Local fLine:Int=True
		Local LX1:Int,LY1:Int,LX2:Int,LY2:Int
		
		Local X2:Int=X+CellWidth,Y2:Int=Y+CellHeight
		
		'Values of sample at edges
		Local UL:Float,UR:Float,LR:Float,LL:Float
		UL=Self._SampleFnc(X,Y)
		UR=Self._SampleFnc(X2,Y)
		LR=Self._SampleFnc(X2,Y2)
		LL=Self._SampleFnc(X,Y2)
		
		'Determine cell type
		Local N:Int
		If (LL>Self._Threshold) Then N=N|%1
		If (LR>Self._Threshold) Then N=N|%10
		If (UR>Self._Threshold) Then N=N|%100
		If (UL>Self._Threshold) Then N=N|%1000
		
		Rem
		Interplation at edges:
		Left:
		Int(Float(CellHeight)*((Self._Threshold-UL)/(LL-UL)))
		Right:
		Int(Float(CellHeight)*((Self._Threshold-UR)/(LR-UR)))
		Upper:
		Int(Float(CellWidth)*((Self._Threshold-UL)/(UR-UL)))
		Lower:
		Int(Float(CellWidth)*((Self._Threshold-LL)/(LR-LL)))
		End Rem
		
		Local Width2:Int=CellWidth Shr 1,Height2:Int=CellHeight Shr 1
		
		'Handle cell types
		Select N
		'Contour cuts off corner
		Case 1,14'Lower left corner
			LX1=X
			LY2=Y2
			If (Self._fInterpolate=True)
				LY1=Y+Int(Float(CellHeight)*((Self._Threshold-UL)/(LL-UL)))'Interpolated at left edge
				LX2=X+Int(Float(CellWidth)*((Self._Threshold-LL)/(LR-LL)))'Interpolated at lower edge
			Else
				LY1=Y+Height2
				LX2=X+Width2
			End If
		Case 2,13'Lower right corner
			LY1=Y2
			LX2=X2
			If (Self._fInterpolate=True)
				LX1=X+Int(Float(CellWidth)*((Self._Threshold-LL)/(LR-LL)))'Interpolated at lower edge
				LY2=Y+Int(Float(CellHeight)*((Self._Threshold-UR)/(LR-UR)))'Interpolated at right edge
			Else
				LX1=X+Width2
				LY2=Y+Height2
			End If
		Case 4,11'Upper right corner
			LY1=Y
			LX2=X2
			If (Self._fInterpolate=True)
				LX1=X+Int(Float(CellWidth)*((Self._Threshold-UL)/(UR-UL)))'Interpolated at upper edge
				LY2=Y+Int(Float(CellHeight)*((Self._Threshold-UR)/(LR-UR)))'Interpolated at right edge
			Else
				LX1=X+Width2
				LY2=Y+Height2
			End If
		Case 7,8'Upper left corner
			LX1=X
			LY2=Y
			If (Self._fInterpolate=True)
				LY1=Y+Int(Float(CellHeight)*((Self._Threshold-UL)/(LL-UL)))'Interpolated at left edge
				LX2=X+Int(Float(CellWidth)*((Self._Threshold-UL)/(UR-UL)))'Interpolated at upper edge
			Else
				LY1=Y+Height2
				LX2=X+Width2
			End If
		'Contour crosses cell at opposing edges
		Case 3,12'horizontally
			LX1=X
			LX2=X2
			If (Self._fInterpolate=True)
				LY1=Y+Int(Float(CellHeight)*((Self._Threshold-UL)/(LL-UL)))'Interpolated at left edge
				LY2=Y+Int(Float(CellHeight)*((Self._Threshold-UR)/(LR-UR)))'Interpolated at right edge
			Else
				LY1=Y+Height2
				LY2=LY1
			End If
		Case 6,9'vertically
			LY1=Y
			LY2=Y2
			If (Self._fInterpolate=True)
				LX1=X+Int(Float(CellWidth)*((Self._Threshold-UL)/(UR-UL)))'Interpolated at upper edge
				LX2=X+Int(Float(CellWidth)*((Self._Threshold-LL)/(LR-LL)))'Interpolated at lower edge
			Else
				LX1=X+Width2
				LX2=LX1
			End If
		'Ambiguous case
		Case 5,10
			'Subdivide
			Local XAdd:Int=X+Width2,YAdd:Int=Y+Height2
			'First subcell: Upper left
			Self.__MarchCell(X,Y,Width2,Height2)
			'Second subcell: Upper right
			Self.__MarchCell(XAdd,Y,Width2,Height2)
			'Third subcell: Lower left
			Self.__MarchCell(X,YAdd,Width2,Height2)
			'Fourth subcell: Lower right
			Self.__MarchCell(XAdd,YAdd,Width2,Height2)
			'Disable line drawing
			fLine=False
		'No contour in cell
		Case 0,15
			'Disable line drawing
			fLine=False
		End Select
		If (fLine=True)'Draw line
			Self._LineFnc(Self.___Target,LX1,LY1,LX2,LY2,Self._ColorARGB)
		End If
	End Method
	
End Type
